#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 1A.2 - Modules, fichiers, expressions régulières
# 
# Le langage Python est défini par un ensemble de règle, une grammaire. Seul, il n'est bon qu'à faire des calculs. Les modules sont des collections de fonctionnalités pour interagir avec des capteurs ou des écrans ou pour faire des calculs plus rapides ou plus complexes.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ### Fichiers
# 
# Les fichiers permettent deux usages principaux :
# 
# - récupérer des données d'une exécution du programme à l'autre (lorsque le programme s'arrête, toutes les variables sont perdues)
# - échanger des informations avec d'autres programmes (Excel par exemple).
# 
# Le format le plus souvent utilisé est le fichier plat, texte, txt, csv, tsv. C'est un fichier qui contient une information structurée sous forme de matrice, en ligne et colonne car c'est comme que les informations numériques se présentent le plus souvent. Un fichier est une longue séquence de caractères. Il a fallu choisir une convention pour dire que deux ensembles de caractères ne font pas partie de la même colonne ou de la même ligne. La convention la plus répandue est :
# 
# * ``\t`` : séparateur de colonnes
# * ``\n`` : séparateur de lignes
# 
# Le caractère ``\`` indique au langage python que le caractère qui suit fait partie d'un code. Vous trouverez la liste des codes : [String and Bytes literals](https://docs.python.org/3.4/reference/lexical_analysis.html#string-and-bytes-literals).
# 
# **Aparté :** aujourd'hui, lire et écrire des fichiers est tellement fréquent qu'il existe des outils qui font ça dans une grande variété de formats. Vous découvrirez cela lors de la [séance 10](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/notebooks/td1a_cenonce_session_10.html). Il est utile pourtant de le faire au moins une fois soi-même pour comprendre la logique des outils et pour ne pas être bloqué dans les cas non prévus.
# 
# Ecrire et lire des fichiers est beaucoup plus long que de jouer avec des variables. Ecrire signifie qu'on enregistre les données sur le disque dur : elles passent du programme au disque dur (elles deviennent permanentes). Elles font le chemin inverse lors de la lecture.

# #### Ecriture
# 
# Il est important de retenir qu'un fichier texte ne peut recevoir que des chaînes de caractères.

# In[2]:


mat = [[1.0, 0.0],[0.0,1.0] ]                    # matrice de type liste de listes
with open ("mat.txt", "w") as f :                # création d'un fichier en mode écriture
    for i in range (0,len (mat)) :               # 
        for j in range (0, len (mat [i])) :      # 
            s = str (mat [i][j])                 # conversion en chaîne de caractères
            f.write (s + "\t")                   #
        f.write ("\n")                           # 
        
# on vérifie que le fichier existe : 
import os
print([ _ for _ in os.listdir(".") if "mat" in _ ] )

# la ligne précédente utilise le symbole _ : c'est une variable 
# le caractère _ est une lettre comme une autre
# on pourrait écrire :
# print([ fichier for fichier in os.listdir(".") if "mat" in fichier ] )
# on utilise cette convention pour dire que cette variable n'a pas vocation à rester


# Le même programme mais écrit avec une écriture condensée :

# In[3]:


mat = [[1.0, 0.0],[0.0,1.0] ]             # matrice de type liste de listes
with open ("mat.txt", "w") as f :        # création d'un fichier
    s = '\n'.join ( '\t'.join( str(x) for x in row  ) for row in mat )
    f.write ( s )
    
# on vérifie que le fichier existe : 
print([ _ for _ in os.listdir(".") if "mat" in _ ] )    


# On regare les premières lignes du fichier ``mat2.txt`` :

# In[4]:


import pyensae
get_ipython().run_line_magic('load_ext', 'pyensae')
get_ipython().run_line_magic('head', 'mat.txt')


# #### Lecture

# In[5]:


with open ("mat.txt", "r") as f : # ouverture d'un fichier
    mat = [ row.strip(' \n').split('\t') for row in f.readlines() ]
print(mat)


# On retrouve les mêmes informations à ceci près qu'il ne faut pas oublier de convertir les nombres initiaux en ``float``.

# In[6]:


with open ("mat.txt", "r") as f :                # ouverture d'un fichier
    mat = [ [ float(x) for x in row.strip(' \n').split('\t') ] for row in f.readlines() ]
print(mat)


# Voilà qui est mieux. Le module [os.path](https://docs.python.org/3.4/library/os.path.html) propose différentes fonctions pour manipuler les noms de fichiers. Le module [os](http://docs.python.org/3.3/library/os.html#module-os) propose différentes fonctions pour manipuler les fichiers :

# In[7]:


import os
for f in os.listdir('.'):
    print (f)


# #### with
# 
# De façon pragmatique, l'instruction [with](https://docs.python.org/3.4/reference/compound_stmts.html#with) permet d'écrire un code plus court d'une instruction : ``close``. Les deux bouts de code suivant sont équivalents :

# In[8]:


with open("exemple_fichier.txt", "w") as f:
    f.write("something")


# In[9]:


f = open("exemple_fichier.txt", "w")
f.write("something")
f.close()


# L'instruction ``close`` *ferme* le fichier. A l'ouverture, le fichier est réservé par le programme Python, aucune autre application ne peut écrire dans le même fichier. Après l'instruction ``close``, une autre application pour le supprimer, le modifier. Avec le mot clé ``with``, la méthode ``close`` est implicitement appelée.

# #### à quoi ça sert ?
# 
# On écrit très rarement un fichier texte. Ce format est le seul reconnu par toutes les applications. Tous les logiciels, tous les langages proposent des fonctionnalités qui exportent les données dans un format texte. Dans certaines circonstances, les outils standards ne fonctionnent pas - trop grops volumes de données, problème d'encoding, caractère inattendu -. Il faut se débrouiller.

# In[10]:





# ### Exercice 1 : Excel $\rightarrow$ Python $\rightarrow$ Excel
#     
# Il faut télécharger le fichier [seance4_excel.xlsx](http://www.xavierdupre.fr/enseignement/complements/seance4_excel.xlsx) qui contient une table de trois colonnes. Il faut :
# 
# - enregistrer le fichier au format texte,
# - le lire sous python
# - créer une matrice carrée 3x3 où chaque valeur est dans sa case (X,Y),
# - enregistrer le résultat sous format texte,
# - le récupérer sous Excel.    

# ### Autres formats de fichiers
# 
# Les fichiers texte sont les plus simples à manipuler mais il existe d'autres formats classiques~:
# 
# - [html](http://fr.wikipedia.org/wiki/Hypertext_Markup_Language) : les pages web
# - [xml](http://fr.wikipedia.org/wiki/Extensible_Markup_Language) : données structurées
# - [zip](http://fr.wikipedia.org/wiki/ZIP_(format_de_fichier), [gz](http://fr.wikipedia.org/wiki/Gzip) : données compressées
# - [wav](http://fr.wikipedia.org/wiki/WAVEform_audio_format), [mp3](http://fr.wikipedia.org/wiki/MPEG-1/2_Audio_Layer_3), [ogg](http://fr.wikipedia.org/wiki/Ogg) : musique
# - [mp4](http://fr.wikipedia.org/wiki/MPEG-4), [Vorbis](http://fr.wikipedia.org/wiki/Vorbis) : vidéo
# - ...

# ### Modules
# 
# Les modules sont des extensions du langages. Python ne sait quasiment rien faire seul mais il bénéficie de nombreuses extensions. On distingue souvent les extensions présentes lors de l'installation du langage (le module [math](https://docs.python.org/3.4/library/math.html)) des extensions externes qu'il faut soi-même installer ([numpy](http://www.numpy.org/)). Deux liens :
# 
# - [modules officiels](http://docs.python.org/3.4/library/)
# - [modules externes](https://pypi.python.org/pypi)
# 
# Le premier réflexe est toujours de regarder si un module ne pourrait pas vous être utile avant de commencer à programmer. Pour utiliser une fonction d'un module, on utilise l'une des syntaxes suivantes :

# In[11]:


import math
print (math.cos(1))

from math import cos
print (cos(1))

from math import *    # cette syntaxe est déconseillée car il est possible qu'une fonction
print (cos(1))        # porte le même nom qu'une des vôtres


# ### Exercice 2 : trouver un module (1)
# 
# Aller à la page [modules officiels](http://docs.python.org/3.4/library/) (ou utiliser un moteur de recherche) pour trouver un module permettant de générer des nombres aléatoires. Créer une liste de nombres aléatoires selon une loi uniforme puis faire une permutation aléatoire de cette séquence.

# ### Exercice 3 : trouver un module (2)
# 
# Trouver un module qui vous permette de calculer la différence entre deux dates puis déterminer le jour de la semaine où vous êtes nés.

# ### Module qu'on crée soi-même
# 
# Il est possible de répartir son programme en plusieurs fichiers. Par exemple, un premier fichier ``monmodule.py`` qui contient une fonction :

# In[12]:


# fichier monmodule.py
import math

def fonction_cos_sequence(seq) :
    return [ math.cos(x) for x in seq ]

if __name__ == "__main__" :
    print ("ce message n'apparaît que si ce programme est le point d'entrée")


# La cellule suivante vous permet d'enregistrer le contenu de la cellule précédente dans un fichier appelée ``monmodule.py``.

# In[13]:


code = """
# -*- coding: utf-8 -*-
import math
def fonction_cos_sequence(seq) :
    return [ math.cos(x) for x in seq ]        
if __name__ == "__main__" :
    print ("ce message n'apparaît que si ce programme est le point d'entrée")
"""
with open("monmodule.py", "w", encoding="utf8") as f :
    f.write(code)


# Le second fichier :

# In[14]:


import monmodule

print ( monmodule.fonction_cos_sequence ( [ 1, 2, 3 ] ) )


# **Note :** Si le fichier ``monmodule.py`` est modifié, *python* ne recharge pas automatiquement le module si celui-ci a déjà été chargé. On peut voir la liste des modules en mémoire dans la variable ``sys.modules`` :

# In[15]:


import sys
list(sorted(sys.modules))[:10]


# Pour retirer le module de la mémoire, il faut l'enlever de ``sys.modules`` avec l'instruction ``del sys.modules['monmodule']``. *Python* aura l'impression que le module *monmodule.py* est nouveau et il l'importera à nouveau.

# ### Exercice 4 : son propre module
# 
# Que se passe-t-il si vous remplacez ``if __name__ == "__main__":``  par ``if True :``, ce qui équivaut à retirer la ligne ``if __name__ == "__main__":`` ?

# ### Expressions régulières
# 
# Pour la suite de la séance, on utilise comme préambule les instructions suivantes :

# In[16]:


import pyensae.datasource
discours = pyensae.datasource.download_data('voeux.zip', website = 'xd')


# La documentation pour les expressions régulières est ici : [regular expressions](http://docs.python.org/3.3/library/re.html). Elles permettent de rechercher des motifs dans un texte :
# 
# - _4 chiffres / 2 chiffres / 2 chiffres_ correspond au motif des dates, avec une expression régulière, il s'écrit : ``[0-9]{4}/[0-9]{2}/[0-9]{2}``
# - _la lettre a répété entre 2 et 10 fois_ est un autre motif, il s'écrit : ``a{2,10}``.

# In[17]:


import re   # les expressions régulières sont accessibles via le module re
expression = re.compile("[0-9]{2}/[0-9]{2}/[0-9]{4}")
texte = """Je suis né le 28/12/1903 et je suis mort le 08/02/1957. Ma seconde femme est morte le 10/11/63."""
cherche = expression.findall(texte)
print(cherche)


# Pourquoi la troisième date n'apparaît pas dans la liste de résultats ?

# ### Exercice 5 : chercher un motif dans un texte
# 
# On souhaite obtenir toutes les séquences de lettres commençant par _je_ ? Quel est le motif correspondant ? Il ne reste plus qu'à terminer le programme précédent.

# In[18]:





# ### Exercice 6 : chercher un autre motif dans un texte
# 
# Avec la même expression régulière, rechercher indifféremment le mot _securite_ ou _insecurite_.

# In[19]:





# On peut passer du temps à construire des expressions assez complexes surtout quand on oublie quelques [Petites subtilités avec les expressions régulières en Python](http://www.xavierdupre.fr/blog/2014-12-02_nojs.html).

# ## Exercice 7 : recherche les urls dans une page wikipédia
# 
# On pourra prendre comme exemple la page du programme [Python](https://fr.wikipedia.org/wiki/Python_(langage)).

# ## Exercice 8 : construire un texte à motif
# 
# A l'inverse des expressions régulières, des modules comme [Mako](http://www.makotemplates.org/) ou [Jinja2](http://jinja.pocoo.org/docs/dev/) permettent de construire simplement des documents qui suivent des règles. Ces outils sont très utilisés pour la construction de page web. On appelle cela faire du [templating](https://wiki.python.org/moin/Templating). Créer une page web qui affiche à l'aide d'un des modules la liste des dimanches de cette année.

# In[20]: